package com.rp.util.sort;

import com.rp.util.sort.ISort;
import com.rp.util.sort.SortException;
import com.rp.util.ReflectionUtil;
import java.util.Comparator;
import java.util.List;

public final class MergeInsertionSort implements ISort
{
	public final static int k = 30;
	private ISort iSort = null;
	
	public MergeInsertionSort ()
	{
		this.iSort = new InsertionSort();	
	}

	public void sort (int[] arr)
	{
		sort (arr, 0, arr.length-1);
	}

	public void sort (int[] arr, int begin, int end)
	{
		int diff = (end - begin);
		if ( diff <= k)
		{
			iSort.sort(arr, begin, end);
			return;	
		}
		
		if (begin < end)
		{
			int mid = (begin+end)/2; 
			sort (arr,begin,mid);
			sort (arr,mid+1,end);
			merge (arr, begin, mid, end);
		}	
	}
	
	protected void merge (int[] arr, int begin, int mid, int end)
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		int[] left = new int[n1];
		int[] right = new int[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr[begin+i];	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr[mid+j+1];	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (left[i] <= right[j])
					{
						arr[k] = left[i];
						i++;	
					}
					else
					{
						arr[k] = right[j];
						j++;
					}
				}
				else
				{
					arr[k] = left[i];
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr[k] = right[j];
					j++;	
				}
			}
		}	
	}	
	
	public void sort (String[] arr)
	{
		sort (arr, 0, arr.length-1);	
	}
	
	public void sort (String[] arr, int begin, int end)
	{
		int diff = (end - begin);
		if ( diff <= k)
		{
			iSort.sort(arr, begin, end);
			return;	
		}
		
		if (begin < end)
		{
			int mid = (begin+end)/2;
			sort (arr, begin, mid);
			sort (arr, mid+1, end);
			merge (arr, begin, mid, end);	
		}	
	}
	
	protected void merge (String[] arr, int begin, int mid, int end)
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		String[] left = new String[n1];
		String[] right = new String[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr[begin+i];	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr[mid+j+1];	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (left[i].compareToIgnoreCase(right[j]) <= 0)
					{
						arr[k] = left[i];
						i++;	
					}
					else
					{
						arr[k] = right[j];
						j++;
					}
				}
				else
				{
					arr[k] = left[i];
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr[k] = right[j];
					j++;	
				}
			}
		}	
	}	

	public void sort (long[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (long[] arr, int begin, int end)
	{
		int diff = (end - begin);
		if ( diff <= k)
		{
			iSort.sort(arr, begin, end);
			return;	
		}
		
		if (begin < end)
		{
			int mid = (begin+end)/2; 
			sort (arr,begin,mid);
			sort (arr,mid+1,end);
			merge (arr, begin, mid, end);
		}	
	}
	
	protected void merge (long[] arr, int begin, int mid, int end)
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		long[] left = new long[n1];
		long[] right = new long[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr[begin+i];	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr[mid+j+1];	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (left[i] <= right[j])
					{
						arr[k] = left[i];
						i++;	
					}
					else
					{
						arr[k] = right[j];
						j++;
					}
				}
				else
				{
					arr[k] = left[i];
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr[k] = right[j];
					j++;	
				}
			}
		}	
	}	

	public void sort (short[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (short[] arr, int begin, int end)
	{
		int diff = (end - begin);
		if ( diff <= k)
		{
			iSort.sort(arr, begin, end);
			return;	
		}
		
		if (begin < end)
		{
			int mid = (begin+end)/2; 
			sort (arr,begin,mid);
			sort (arr,mid+1,end);
			merge (arr, begin, mid, end);
		}	
	}
	
	protected void merge (short[] arr, int begin, int mid, int end)
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		short[] left = new short[n1];
		short[] right = new short[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr[begin+i];	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr[mid+j+1];	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (left[i] <= right[j])
					{
						arr[k] = left[i];
						i++;	
					}
					else
					{
						arr[k] = right[j];
						j++;
					}
				}
				else
				{
					arr[k] = left[i];
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr[k] = right[j];
					j++;	
				}
			}
		}	
	}	
	
	public void sort (double[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (double[] arr, int begin, int end)
	{
		int diff = (end - begin);
		if ( diff <= k)
		{
			iSort.sort(arr, begin, end);
			return;	
		}
		
		if (begin < end)
		{
			int mid = (begin+end)/2; 
			sort (arr,begin,mid);
			sort (arr,mid+1,end);
			merge (arr, begin, mid, end);
		}	
	}
	
	protected void merge (double[] arr, int begin, int mid, int end)
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		double[] left = new double[n1];
		double[] right = new double[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr[begin+i];	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr[mid+j+1];	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (left[i] <= right[j])
					{
						arr[k] = left[i];
						i++;	
					}
					else
					{
						arr[k] = right[j];
						j++;
					}
				}
				else
				{
					arr[k] = left[i];
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr[k] = right[j];
					j++;	
				}
			}
		}	
	}	

	public void sort (float[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (float[] arr, int begin, int end)
	{
		int diff = (end - begin);
		if ( diff <= k)
		{
			iSort.sort(arr, begin, end);
			return;	
		}
		
		if (begin < end)
		{
			int mid = (begin+end)/2; 
			sort (arr,begin,mid);
			sort (arr,mid+1,end);
			merge (arr, begin, mid, end);
		}	
	}
	
	protected void merge (float[] arr, int begin, int mid, int end)
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		float[] left = new float[n1];
		float[] right = new float[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr[begin+i];	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr[mid+j+1];	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (left[i] <= right[j])
					{
						arr[k] = left[i];
						i++;	
					}
					else
					{
						arr[k] = right[j];
						j++;
					}
				}
				else
				{
					arr[k] = left[i];
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr[k] = right[j];
					j++;	
				}
			}
		}	
	}	

	public void sort (char[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (char[] arr, int begin, int end)
	{
		int diff = (end - begin);
		if ( diff <= k)
		{
			iSort.sort(arr, begin, end);
			return;	
		}
		
		if (begin < end)
		{
			int mid = (begin+end)/2; 
			sort (arr,begin,mid);
			sort (arr,mid+1,end);
			merge (arr, begin, mid, end);
		}	
	}
	
	protected void merge (char[] arr, int begin, int mid, int end)
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		char[] left = new char[n1];
		char[] right = new char[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr[begin+i];	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr[mid+j+1];	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (left[i] <= right[j])
					{
						arr[k] = left[i];
						i++;	
					}
					else
					{
						arr[k] = right[j];
						j++;
					}
				}
				else
				{
					arr[k] = left[i];
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr[k] = right[j];
					j++;	
				}
			}
		}	
	}	
	
	public void sort (byte[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (byte[] arr, int begin, int end)
	{
		int diff = (end - begin);
		if ( diff <= k)
		{
			iSort.sort(arr, begin, end);
			return;	
		}
		
		if (begin < end)
		{
			int mid = (begin+end)/2; 
			sort (arr,begin,mid);
			sort (arr,mid+1,end);
			merge (arr, begin, mid, end);
		}	
	}
	
	protected void merge (byte[] arr, int begin, int mid, int end)
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		byte[] left = new byte[n1];
		byte[] right = new byte[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr[begin+i];	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr[mid+j+1];	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (left[i] <= right[j])
					{
						arr[k] = left[i];
						i++;	
					}
					else
					{
						arr[k] = right[j];
						j++;
					}
				}
				else
				{
					arr[k] = left[i];
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr[k] = right[j];
					j++;	
				}
			}
		}	
	}	
	
	public void sort (Object[] arr, Comparator c, String methodName) throws SortException
	{
		sort (arr, c, methodName, new Class[0], new Object[0]);	
	}
	
	public void sort (Object[] arr, Comparator c, String methodName, Class[] paramTypes, Object[] paramValues)  
							throws SortException
	{
		sort (arr, 0, arr.length-1, c, methodName, paramTypes, paramValues);	
	}
	
	public void sort (Object[] arr, int begin, int end, Comparator c, String methodName)
							throws SortException
	{
		sort (arr, begin, end, c, methodName, new Class[0], new Object[0]);
	}
	
	public void sort (Object[] arr, int begin, int end, Comparator c, String methodName, Class[] paramTypes,
								Object[] paramValues) throws SortException
	{
		try
		{
			int diff = (end - begin);
			if ( diff <= k)
			{
				iSort.sort(arr, begin, end, c, methodName, paramTypes, paramValues);
				return;	
			}
			
			if (begin < end)
			{
				int mid = (begin+end)/2;
				sort (arr, begin, mid, c, methodName, paramTypes, paramValues);
				sort (arr, mid+1, end, c, methodName, paramTypes, paramValues);
				merge (arr, c, methodName, paramTypes, paramValues, begin, mid, end);	
			}	
		}
		catch (Exception e)
		{
			throw new SortException (e);	
		}
	}
	
	protected void merge (Object[] arr, Comparator c, String methodName, Class[] paramTypes,
							Object[] paramValues, int begin, int mid, int end) throws Exception
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		Object[] left = new Object[n1];
		Object[] right = new Object[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr[begin+i];	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr[mid+j+1];	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (c.compare (ReflectionUtil.getMethodValue (left[i], methodName, paramTypes, paramValues), 
						ReflectionUtil.getMethodValue (right[j], methodName, paramTypes, paramValues)) <= 0 )
					{
						arr[k] = left[i];
						i++;	
					}
					else
					{
						arr[k] = right[j];
						j++;
					}
				}
				else
				{
					arr[k] = left[i];
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr[k] = right[j];
					j++;	
				}
			}
		}	
	}	
	
	public void sort (Object[] arr, Comparator c)
	{
		sort (arr, 0, arr.length-1, c);	
	}
	
	public void sort (Object[] arr, int begin, int end, Comparator c)
	{
		int diff = (end - begin);
		if ( diff <= k)
		{
			iSort.sort(arr, begin, end, c);
			return;	
		}
		
		if (begin < end)
		{
			int mid = (begin+end)/2;
			sort (arr, begin, mid, c);
			sort (arr, mid+1, end, c);
			merge (arr, c, begin, mid, end);	
		}
	}
	
	protected void merge (Object[] arr, Comparator c, int begin, int mid, int end)
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		Object[] left = new Object[n1];
		Object[] right = new Object[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr[begin+i];	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr[mid+j+1];	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (c.compare(left[i], right[j]) <= 0)
					{
						arr[k] = left[i];
						i++;	
					}
					else
					{
						arr[k] = right[j];
						j++;
					}
				}
				else
				{
					arr[k] = left[i];
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr[k] = right[j];
					j++;	
				}
			}
		}	
	}	
	
	public void sort (List arr, Comparator c)
	{
		sort (arr, 0, arr.size()-1, c);	
	}
	
	public void sort (List arr, int begin, int end, Comparator c)
	{
		int diff = (end - begin);
		if ( diff <= k)
		{
			iSort.sort(arr, begin, end, c);
			return;	
		}
		
		if (begin < end)
		{
			int mid = (begin+end)/2;
			sort (arr, begin, mid, c);
			sort (arr, mid+1, end, c);
			merge (arr, c, begin, mid, end);	
		}
	}
	
	protected void merge (List arr, Comparator c, int begin, int mid, int end)
	{
		int n1 = mid - begin + 1;
		int n2 = end - mid;
		Object[] left = new Object[n1];
		Object[] right = new Object[n2];
		
		for (int i = 0; i < n1; i++)
		{
			left[i] = arr.get(begin+i);	
		}	
		
		for (int j = 0; j < n2; j++)
		{
			right[j] = arr.get(mid+j+1);	
		}
		
		int i = 0;
		int j = 0;
		int leftSize = left.length;
		int rightSize = right.length;
		for (int k = begin; k <= end; k++)
		{
			if (i < leftSize)
			{
				if (j < rightSize)
				{
					if (c.compare(left[i], right[j]) <= 0)
					{
						arr.set(k, left[i]);
						i++;	
					}
					else
					{
						arr.set(k, right[j]);
						j++;
					}
				}
				else
				{
					arr.set(k, left[i]);
					i++;	
				}
			}
			else
			{
				if (j < rightSize)
				{
					arr.set(k, right[j]);
					j++;	
				}
			}
		}	
	}	
}