package com.rp.util.sort;

import com.rp.util.sort.ISort;
import com.rp.util.sort.SortException;
import com.rp.util.ReflectionUtil;
import java.util.Comparator;
import java.util.Random;
import java.util.List;

public final class QuickSortRandom implements ISort
{
	private Random random = null;
	
	public QuickSortRandom ()
	{
		random = new Random();	
	}
	
	public void sort (int[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (int[] arr, int begin, int end)
	{

		if (end > begin)
		{
			int low = begin;
			int high = end;
			int middle = arr[getMiddle (begin, end)];
			int temp;
			while (low <= high)
			{
				while (low < end && arr[low] < middle)
				{
					low++;
				}
				while (high > begin && arr[high] > middle)
				{
					high--;	
				}
				if (low <= high)
				{
					temp = arr[low];
					arr[low] = arr[high];
					arr[high] = temp;
					low++;
					high--;	
				}
			}
			if (begin < high)
			{
				sort (arr, begin, high);	
			}
			if (low < end)
			{
				sort (arr, low, end);
			}
		}
	}

	public void sort (String[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (String[] arr, int begin, int end)
	{
		if (end > begin)
		{
			int low = begin;
			int high = end;
			String middle = arr[getMiddle (begin, end)];
			String temp;
			while (low <= high)
			{
				while (low < end && arr[low].compareToIgnoreCase(middle) < 0)
				{
					low++;
				}
				while (high > begin && arr[high].compareToIgnoreCase(middle) > 0)
				{
					high--;	
				}
				if (low <= high)
				{
					temp = arr[low];
					arr[low] = arr[high];
					arr[high] = temp;
					low++;
					high--;	
				}
			}
			if (begin < high)
			{
				sort (arr, begin, high);	
			}
			if (low < end)
			{
				sort (arr, low, end);
			}
		}
	}
	
	public void sort (long[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (long[] arr, int begin, int end)
	{
		if (end > begin)
		{
			int low = begin;
			int high = end;
			long middle = arr[getMiddle (begin, end)];
			long temp;
			while (low <= high)
			{
				while (low < end && arr[low] < middle)
				{
					low++;
				}
				while (high > begin && arr[high] > middle)
				{
					high--;	
				}
				if (low <= high)
				{
					temp = arr[low];
					arr[low] = arr[high];
					arr[high] = temp;
					low++;
					high--;	
				}
			}
			if (begin < high)
			{
				sort (arr, begin, high);	
			}
			if (low < end)
			{
				sort (arr, low, end);
			}
		}
	}
	
	public void sort (short[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (short[] arr, int begin, int end)
	{
		if (end > begin)
		{
			int low = begin;
			int high = end;
			short middle = arr[getMiddle (begin, end)];
			short temp;
			while (low <= high)
			{
				while (low < end && arr[low] < middle)
				{
					low++;
				}
				while (high > begin && arr[high] > middle)
				{
					high--;	
				}
				if (low <= high)
				{
					temp = arr[low];
					arr[low] = arr[high];
					arr[high] = temp;
					low++;
					high--;	
				}
			}
			if (begin < high)
			{
				sort (arr, begin, high);	
			}
			if (low < end)
			{
				sort (arr, low, end);
			}
		}
	}
	
	public void sort (double[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (double[] arr, int begin, int end)
	{
		if (end > begin)
		{
			int low = begin;
			int high = end;
			double middle = arr[getMiddle (begin, end)];
			double temp;
			while (low <= high)
			{
				while (low < end && arr[low] < middle)
				{
					low++;
				}
				while (high > begin && arr[high] > middle)
				{
					high--;	
				}
				if (low <= high)
				{
					temp = arr[low];
					arr[low] = arr[high];
					arr[high] = temp;
					low++;
					high--;	
				}
			}
			if (begin < high)
			{
				sort (arr, begin, high);	
			}
			if (low < end)
			{
				sort (arr, low, end);
			}
		}
	}
	
	protected void swap (double[] arr, int low, int high)
	{
		double elem = arr[low];
		arr[low] = arr[high];
		arr[high] = elem;	
	}

	public void sort (float[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (float[] arr, int begin, int end)
	{
		if (end > begin)
		{
			int low = begin;
			int high = end;
			float middle = arr[getMiddle (begin, end)];
			float temp;
			while (low <= high)
			{
				while (low < end && arr[low] < middle)
				{
					low++;
				}
				while (high > begin && arr[high] > middle)
				{
					high--;	
				}
				if (low <= high)
				{
					temp = arr[low];
					arr[low] = arr[high];
					arr[high] = temp;
					low++;
					high--;	
				}
			}
			if (begin < high)
			{
				sort (arr, begin, high);	
			}
			if (low < end)
			{
				sort (arr, low, end);
			}
		}
	}
	
	protected void swap (float[] arr, int low, int high)
	{
		float elem = arr[low];
		arr[low] = arr[high];
		arr[high] = elem;	
	}

	public void sort (char[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (char[] arr, int begin, int end)
	{
		if (end > begin)
		{
			int low = begin;
			int high = end;
			char middle = arr[getMiddle (begin, end)];
			char temp;
			while (low <= high)
			{
				while (low < end && arr[low] < middle)
				{
					low++;
				}
				while (high > begin && arr[high] > middle)
				{
					high--;	
				}
				if (low <= high)
				{
					temp = arr[low];
					arr[low] = arr[high];
					arr[high] = temp;
					low++;
					high--;	
				}
			}
			if (begin < high)
			{
				sort (arr, begin, high);	
			}
			if (low < end)
			{
				sort (arr, low, end);
			}
		}
	}
	
	public void sort (byte[] arr)
	{
		sort (arr, 0, arr.length-1);
	}
	
	public void sort (byte[] arr, int begin, int end)
	{
		if (end > begin)
		{
			int low = begin;
			int high = end;
			byte middle = arr[getMiddle (begin, end)];
			byte temp;
			while (low <= high)
			{
				while (low < end && arr[low] < middle)
				{
					low++;
				}
				while (high > begin && arr[high] > middle)
				{
					high--;	
				}
				if (low <= high)
				{
					temp = arr[low];
					arr[low] = arr[high];
					arr[high] = temp;
					low++;
					high--;	
				}
			}
			if (begin < high)
			{
				sort (arr, begin, high);	
			}
			if (low < end)
			{
				sort (arr, low, end);
			}
		}
	}

	public void sort (Object[] arr, Comparator c, String methodName) throws SortException
	{
		sort (arr, c, methodName, new Class[0], new Object[0]);	
	}
	
	public void sort (Object[] arr, Comparator c, String methodName, Class[] paramTypes, Object[] paramValues)
						 throws SortException
	{
		sort (arr, 0, arr.length-1, c, methodName, paramTypes, paramValues);
	}
	
	public void sort (Object[] arr, int begin, int end, Comparator c, String methodName) 
							throws SortException
	{
		sort (arr, begin, end, c, methodName, new Class[0], new Object[0]);
	}

	public void sort (Object[] arr, int begin, int end, Comparator c, String methodName, Class[] paramTypes,
								Object[] paramValues) throws SortException
	{
		try
		{
			if (end > begin)
			{
				int low = begin;
				int high = end;
				Object middle = ReflectionUtil.getMethodValue(arr[getMiddle (begin,end)], methodName, paramTypes, paramValues);
				Object temp;
				while (low <= high)
				{
					while (low < end && c.compare(ReflectionUtil.getMethodValue(arr[low], methodName, paramTypes, paramValues),
							 middle) < 0)
					{
						low++;
					}
					while (high > begin && c.compare(ReflectionUtil.getMethodValue (arr[high], methodName, paramTypes, paramValues), 
							 middle) > 0)
					{
						high--;	
					}
					if (low <= high)
					{
						temp = arr[low];
						arr[low] = arr[high];
						arr[high] = temp;
						low++;
						high--;	
					}
				}
				if (begin < high)
				{
					sort (arr, begin, high, c, methodName, paramTypes, paramValues);	
				}
				if (low < end)
				{
					sort (arr, low, end, c, methodName, paramTypes, paramValues);
				}
			}
		}
		catch (Exception e)
		{
			throw new SortException (e);	
		}
	}
	
	public void sort (Object[] arr, Comparator c)
	{
		sort (arr, 0, arr.length-1, c);
	}
	
	public void sort (Object[] arr, int begin, int end, Comparator c)
	{
		if (end > begin)
		{
			int low = begin;
			int high = end;
			Object middle = arr[getMiddle (begin, end)];
			Object temp;
			while (low <= high)
			{
				while (low < end && c.compare(arr[low], middle) < 0)
				{
					low++;
				}
				while (high > begin && c.compare(arr[high], middle) > 0)
				{
					high--;	
				}
				if (low <= high)
				{
					temp = arr[low];
					arr[low] = arr[high];
					arr[high] = temp;
					low++;
					high--;	
				}
			}
			if (begin < high)
			{
				sort (arr, begin, high, c);	
			}
			if (low < end)
			{
				sort (arr, low, end, c);
			}
		}
	}
	
	public void sort (List arr, Comparator c)
	{
		sort (arr, 0, arr.size()-1, c);
	}
	
	public void sort (List arr, int begin, int end, Comparator c)
	{
		if (end > begin)
		{
			int low = begin;
			int high = end;
			Object middle = arr.get(getMiddle (begin, end));
			Object temp;
			while (low <= high)
			{
				while (low < end && c.compare(arr.get(low), middle) < 0)
				{
					low++;
				}
				while (high > begin && c.compare(arr.get(high), middle) > 0)
				{
					high--;	
				}
				if (low <= high)
				{
					temp = arr.get(low);
					arr.set(low, arr.get(high));
					arr.set(high, temp);
					low++;
					high--;	
				}
			}
			if (begin < high)
			{
				sort (arr, begin, high, c);	
			}
			if (low < end)
			{
				sort (arr, low, end, c);
			}
		}
	}

	protected int getMiddle (int begin, int end)
	{
		int i = this.random.nextInt( (end-begin) );
		return i + begin;
	}
}